import { isWeakable } from "./utils";
export var createStrongStorage = function (startIndex, endIndex, storage) {
    if (startIndex === void 0) { startIndex = 0; }
    if (endIndex === void 0) { endIndex = Infinity; }
    if (storage === void 0) { storage = new WeakMap(); }
    return ({
        get: function (args) {
            var max = Math.min(endIndex, args.length);
            var reads = 0;
            for (var i = startIndex; i < max; ++i) {
                if (isWeakable(args[i])) {
                    reads++;
                    var test_1 = storage.get(args[i]);
                    if (test_1) {
                        if (test_1.arguments.length === max &&
                            test_1.arguments.every(function (v, index) { return v === args[index]; })) {
                            return {
                                value: test_1.storedValue,
                                index: i,
                            };
                        }
                    }
                }
            }
            if (!reads) {
                console.log('arguments given', args);
                throw new Error('No weak-mappable object found to read a cache from. Given [' + args.map(function (a) { return typeof a; }).join(', ') + ']');
            }
            return undefined;
        },
        set: function (args, value) {
            var max = Math.min(endIndex, args.length);
            for (var i = startIndex; i < max; ++i) {
                if (isWeakable(args[i])) {
                    storage.set(args[i], {
                        storedValue: value,
                        arguments: args,
                    });
                }
            }
            return value;
        }
    });
};
